#!/usr/bin/env python3

# SECLAF Sequence Classification Framework
# Copyright (c) 2016-2017 Balazs Szalkai

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os,sys,re,math,time
try:
	import Queue
except:
	import queue as Queue
import numpy as np
import json

from .seq_util import *
from .tree import *
from .annotation import *

# For each class (node) in the tree, returns the number of sequences in the annotation reader that belong to that class.
# Returns: dict: nodeId --> number_of_sequences
def getClassSizes(tree, annotation):
	print("Getting class sizes")
	annotation.rewind()
	classSizes = {n:0 for n in tree.parents}
	counter = 0
	while True:
		seq, classes = annotation.read()
		if seq is None:
			assert classes is None
			break
		counter += 1
		if counter % 1000 == 0:
			sys.stderr.write(".")
			sys.stderr.flush()
		classes = tree.includeParents(classes)
		for c in classes:
			classSizes[c] += 1
	print("Done")
	return classSizes

# Same as getClassSizes(), but loads the class sizes from a cache file if that file exists,
# and calculates and saves the class sizes to that file if it doesn't.
def getClassSizesCached(tree, annotation, cacheFilename):
	try:
		with open(cacheFilename) as f:
			classSizes = json.load(f)
			print("classSizes successfully loaded from cache %s" % cacheFilename)
	except (IOError, OSError):
		assert annotation is not None, "Class sizes cache file %s does not exist." % cacheFilename
		classSizes = getClassSizes(tree, annotation)
		with open(cacheFilename, "w") as f:
			f.write(json.dumps(classSizes))
		print("classSizes successfully saved to cache %s" % cacheFilename)
	return classSizes
		
# Encodes an annotation (which sequence belongs to which family) as a numpy array,
# i.e. produces the target y of the network.
class AnnotationEncoder:
	# tree: a Tree instance
	# maxLevel: maximum level of nodes to consider in the tree (0: root level)
	# minClassSize: minimum number of sequences that have to belong to a considered node
	# classSizes: dictionary of nodeId --> classSize (produced by getClassSizes...())
	def __init__(self, tree, maxLevel, minClassSize, classSizes):
		self.tree = tree
		self.maxLevel = maxLevel # inclusive
		self.minClassSize = minClassSize
		self.classSizes = classSizes
		assert len(self.tree.getRoots()) > 0
		self.buildMapping()
		
	# Builds the mapping from tree nodes (classes) to coordinates.
	# Automatically called on class initialization.
	def buildMapping(self):
		print("AnnotationEncoder: building mapping")
		self.nodeLevel = {}
		self.nodeCoord = {}
		self.coordToNode = []
		
		q = Queue.Queue()
		coordFactory = 0
		for n in self.tree.getRoots():
			if self.minClassSize <= 0 or self.classSizes[n] >= self.minClassSize:
				q.put(n)
				self.nodeLevel[n] = 0
				self.nodeCoord[n] = coordFactory
				self.coordToNode.append(n)
				coordFactory += 1
		while not q.empty():
			n = q.get()
			nLevel = self.nodeLevel[n]
			if nLevel < self.maxLevel:
				for c in self.tree.children[n]:
					if c not in self.nodeLevel and (self.minClassSize <= 0 or self.classSizes[c] >= self.minClassSize):
						q.put(c)
						self.nodeLevel[c] = nLevel+1
						self.nodeCoord[c] = coordFactory
						self.coordToNode.append(c)
						coordFactory += 1
					
		newMaxLevel = np.amax(list(self.nodeLevel.values()))
		assert newMaxLevel <= self.maxLevel, "%s %s" % (newMaxLevel, self.maxLevel)
		self.maxLevel = newMaxLevel
		
		self.dim = coordFactory # dimension of y vectors
		# create characteristic vectors for all the levels
		self.levelPositions = {level: [self.nodeCoord[n] for n in self.nodeLevel if self.nodeLevel[n] == level] for level in range(self.maxLevel+1)}
		self.levelStart = {level: np.amin(self.levelPositions[level]) for level in range(self.maxLevel+1)} # inclusive
		self.levelEnd = {level: np.amax(self.levelPositions[level])+1 for level in range(self.maxLevel+1)} # non-inclusive
		print("Output (annotation) dimensions: %s" % self.dim)
			
	# Returns the number of channels in the resulting encoded array.
	def getDim(self):
		return self.dim
		
	# Encodes a class set ("classes") as a numpy array.
	# In the result, each active class will get 1 weight and the others 0 weight.
	# if eachLevelADistribution is True, then each non-zero level will be a probability distribution instead (weight sum is 1)
	# and if a level is zero, it will be filled with a discrete uniform distribution.
	def encode(self, classes, eachLevelADistribution = False):
		classes = self.tree.includeParents(classes)
		
		if eachLevelADistribution:
			countPerLevel = [0]*(self.maxLevel+1)
			for n in classes:
				if n in self.nodeLevel: # class was large enough and level is small enough
					countPerLevel[self.nodeLevel[n]] += 1

		v = np.zeros(self.getDim())
		for n in classes:
			if n in self.nodeLevel:
				if eachLevelADistribution:
					c = countPerLevel[self.nodeLevel[n]]
					assert c > 0
					v[self.nodeCoord[n]] = 1.0 / c
				else:
					v[self.nodeCoord[n]] = 1.0
					
		if eachLevelADistribution:
			for level in range(self.maxLevel+1):
				if countPerLevel[level] == 0:
					a = self.levelStart[level]
					b = self.levelEnd[level]
					assert b > a
					v[a:b] = 1.0 / (b-a)
			
		return v
		
	# Decodes a numpy array to classes.
	# Returns a tuple of: (classes (maybe more than one per level), confidences (for each class returned)).
	# Parents of a selected node are not necessarily selected.
	def decode(self, v):
		classes = []
		confidences = []
		for n in self.nodeCoord:
			value = v[self.nodeCoord[n]]
			if value >= 0.5:
				classes.append(n)
				confidences.append(value)
		return classes, confidences
