#!/usr/bin/env python3

# SECLAF Sequence Classification Framework
# Copyright (c) 2016-2017 Balazs Szalkai

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os,re,math,time,sys
import numpy as np
import tensorflow as tf

# layers.py: Functions implementing some Tensorflow layers needed by sequence classification networks.

# Parametric rectified linear unit.
# Under the same scope, prelu will reuse the same parameter!
def prelu(x):
	v = tf.get_variable("prelu_weight", dtype=tf.float32, initializer=tf.constant(0.2))
	return tf.where(tf.greater_equal(x, 0), x, x*v)
	
# Like Tensorflow batch_norm, but this one can handle variable length sequences.
# inputs: shape=[batch_size, max_length, channels] or [batch_size, 1, max_length, channels], shape cannot be undef
# lengths: shape=[batch_size], dtype=int32
def sequence_batch_norm(inputs, lengths, **kwargs):
	s_orig = inputs.get_shape().as_list()
	ndim_orig = len(s_orig)
	
	if ndim_orig == 4:
		batch_size, max_length, channels = s_orig[0], s_orig[2], s_orig[3]
	else:
		batch_size, max_length, channels = s_orig[0], s_orig[1], s_orig[2]
	
	mask = tf.cast(tf.sequence_mask(lengths, max_length), inputs.dtype)
	mask = tf.reshape(mask, [batch_size * max_length])
	inputs = tf.reshape(inputs, [batch_size * max_length, channels])
	
	normed = tf.contrib.layers.batch_norm(inputs, batch_weights=mask, **kwargs)
	normed = tf.reshape(normed, s_orig)
	return normed
	
# Spatial Pyramid Pooling in 1D.
# sequences: shape=[batch_size, max_length, channels], dtype=float32
# lengths: shape=[batch_size], dtype=int32
# result: shape=[batch_size, spp_count, channels], dtype=float32
def spp(sequences, lengths, levels, divsPerLevel):
	with tf.name_scope("spp"):
		batch_size, _, channels = sequences.get_shape()
		batchValues = []
		# we cannot max_pool2d because sequences have different lengths
		for b in range(batch_size):
			currLength = lengths[b]
			pooledValues = []
			for level in range(levels):
				ndiv = divsPerLevel**level
				assert ndiv > 0
				divLength = tf.cast(tf.ceil(tf.truediv(currLength, ndiv) - 1e-8), tf.int32)
				for divIndex in range(ndiv):
					divStart = 0 if ndiv <= 1 else tf.cast(tf.round(tf.truediv(currLength-divLength, ndiv-1) * divIndex), tf.int32)
					pooledValues.append(tf.cond(tf.greater(divLength, 0), lambda: tf.reduce_max(sequences[b,divStart:divStart+divLength,:], 0), lambda: tf.zeros(shape=[channels],dtype=tf.float32)))
			spp_count = len(pooledValues)
			pooledValue = tf.stack(pooledValues, 0)
			pooledValue.set_shape([spp_count, channels])
			batchValues.append(pooledValue)
		result = tf.stack(batchValues, 0)
		result.set_shape([batch_size, spp_count, channels])
	return result
	
# Returns total number of floating-point network parameters.
# scope must be a string (scope name), from where weight tensors are taken.
def networkParamCount(scope):
	if not scope.endswith("/"):
		scope += "/"
	vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES)
	paramCount = 0
	for v in vars:
		if v.name.startswith(scope):
			v_size = int(np.prod(v.get_shape().as_list()))
			#print("%s: %s params" % (v.name, v_size))
			paramCount += v_size
	assert paramCount > 0
	return paramCount
