#!/usr/bin/env python3

# SECLAF Sequence Classification Framework
# Copyright (c) 2016-2017 Balazs Szalkai

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os,re,math,time,sys
import numpy as np
import tensorflow as tf

from .layers import *
from . import config
from .utils import *

# network.py: The implementation of the sequence network builder. This class (SeqNetwork) uses config.py for parameters.
# Especially important is config.network, whose structure is described in readme.html.
# config.network describes the layers comprising the classifier network.

# The two stages of the network pipeline.
# spp can be used for transitioning to the Fixed stage, where the network should end.
class NetworkStage:
	VarSeq = 0 # [batch_size, max_length, channels], with currLen defining the actual length for each sequence
	Fixed = 1 # flattened: [batch_size, n] for each sequence

# The implementation of the sequence network builder.
class SeqNetwork:
	# Utility function for generating a scope name and whether it is already used (=should be reused).
	def makeScope(self, type, userScope):
		if userScope is not None:
			if userScope in self.scopes:
				print("Warning: reusing scope %s for %s layer" % (userScope, type))
				return userScope, True
			else:
				self.scopes.append(userScope)
				return userScope, False
		else:
			for i in range(1000000):
				scope = type+str(i)
				if scope not in self.scopes:
					self.scopes.append(scope)
					return scope, False
			assert False
			
	# Parses fully connected layer unit count.
	def parseUnitCount(self, n):
		if n == "#out":
			return self.annEnc.getDim()
		else:
			return int(n)

	# Creates a network based on a sequence encoder, an annotation encoder, and the config module.
	# self.inputSeq: [batch_size, max_length, channels], float32
	# self.inputSeqLen: [batch_size], int32
	# self.yTarget: [batch_size, annEnc.getDim()]
	def __init__(self, seqEncoder, annEnc):
		print("Creating network")
		with tf.variable_scope("seqNetwork") as networkScope:
			self.annEnc = annEnc
			self.inputSeq = tf.placeholder(shape=[config.batchSize, config.maxLength, seqEncoder.getDim()], dtype=tf.float32, name="inputSeq")
			self.inputSeqLen = tf.placeholder(shape=[config.batchSize], dtype=tf.int32, name="inputSeqLen")
			self.yTarget = tf.placeholder(shape=[config.batchSize, annEnc.getDim()], dtype=tf.float32, name="yTarget")
			self.isTraining = tf.placeholder_with_default(input=tf.constant(False), shape=[], name="isTraining")
		
			# determine minimum suitable sequence length for convolutional layers
			minLen = 1 # actual minimum length
			for layer in reversed(config.network):
				type = layer["layer"]
				if type == "maxpool":
					poolSize = layer["k"]
					poolStride = layer.get("s", poolSize)
					minLen = (minLen-1)*poolStride + poolSize
				elif type == "conv":
					convSize = layer["k"]
					convStride = layer.get("s", 1)
					minLen = (minLen-1)*convStride + convSize
			assert config.minLength >= minLen, "config.minLength should be at least %s because of the NN" % minLen
			print("Minimum sequence length for NN is %s" % minLen)
				
			x = tf.expand_dims(self.inputSeq, 1)
			weights_regularizer = tf.contrib.layers.l2_regularizer(scale=1.0)
			self.activation_fn = {"none":None, "prelu":prelu, "relu":tf.nn.relu, "sigmoid":tf.sigmoid, "abs":tf.abs, "tanh":tf.tanh}

			stage = NetworkStage.VarSeq
			self.scopes = []
			self.layerTable = []
			self.layerTable.append("<b>%s</b>" % config.name)
			currLen = self.inputSeqLen
			currMinLen = config.minLength
			
			for layer in config.network:
				type = layer["layer"]
				scope, reuse = self.makeScope(type, layer.get("scope"))
				if type in self.activation_fn:
					x = self.activation_fn[type](x)
					self.layerTable.append(type)
				elif type == "conv":
					assert stage == NetworkStage.VarSeq, "Convolution cannot be used in the fixed stage, only in the variable stage"
					convSize = layer["k"]
					convStride = layer.get("s", 1)
					convDepth = layer["d"] # filter count
					convPadding = "VALID"
					convAct = layer.get("act", "none")
					
					currLen = tf.maximum(0, currLen - (convSize-1))
					currMinLen = max(0, currMinLen - (convSize-1))
					
					x = tf.contrib.layers.convolution2d(x, convDepth, [1, convSize],
						scope=scope, reuse=reuse,
						stride=convStride, padding=convPadding,
						activation_fn=self.activation_fn[convAct], weights_regularizer=weights_regularizer)
						
					self.layerTable.append("conv (size=%s, stride=%s, depth=%s, padding=%s, activation=%s)" % (convSize, convStride, convDepth, convPadding, convAct))
				elif type == "maxpool":
					assert stage == NetworkStage.VarSeq, "Maxpool cannot be used in the fixed stage, only in the variable stage"
					assert not reuse, "Maxpool has no reusable variables"
					poolSize = layer["k"]
					poolStride = layer.get("s", poolSize)
					poolPadding = "VALID"
					
					currLen = tf.floordiv(currLen, poolSize)
					currMinLen = int(currMinLen / poolSize)
					
					x = tf.contrib.layers.max_pool2d(x, kernel_size=[1,poolSize], stride=[1,poolStride], scope=scope, padding=poolPadding)
					
					self.layerTable.append("max_pool (size=%s, stride=%s, padding=%s)" % (poolSize, poolStride, poolPadding))
				elif type == "bn":
					scale = layer.get("scale", 0) != 0
					if stage == NetworkStage.VarSeq:
						x = sequence_batch_norm(x, currLen, scale=scale, is_training = self.isTraining, scope=scope, reuse=reuse)
					elif stage == NetworkStage.Fixed:
						x = tf.contrib.layers.batch_norm(x, scale=scale, is_training = self.isTraining, scope=scope, reuse=reuse)
					else:
						assert False, "Invalid stage "+str(stage)

					self.layerTable.append("batch_norm (scale=%s)" % (scale))
				elif type == "spp":
					assert stage == NetworkStage.VarSeq, "SPP cannot be used in the fixed stage, only in the variable stage"
					assert currMinLen >= 1, "Minimum sequence length %s is too small for neural network!" % config.minLength
					x = tf.squeeze(x, [1])
					levels = layer["levels"]
					divsPerLevel = layer["div"]
					x = spp(x, currLen, levels=levels, divsPerLevel=divsPerLevel)
					sh = x.get_shape().as_list()
					x = tf.reshape(x, [sh[0], sh[1] * sh[2]], name="flatten_after_spp")
					stage = NetworkStage.Fixed
					
					self.layerTable.append("spp (levels=%s, divs_per_level=%s)" % (levels, divsPerLevel))
				elif type == "fc":
					assert stage == NetworkStage.Fixed, "Fully connected layer cannot be used in the variable stage, only in the fixed stage"
					unitCount = self.parseUnitCount(layer["n"])
					activation = layer.get("act", "none")
					x = tf.contrib.layers.fully_connected(x, unitCount,
						activation_fn=self.activation_fn[activation], weights_regularizer=weights_regularizer, scope=scope, reuse=reuse)

					self.layerTable.append("fully_connected (units=%s, activation=%s)" % (unitCount, activation))
				elif type == "dropout":
					assert not reuse, "Dropout has no reusable variables"
					p = layer["p"]
					x = tf.contrib.layers.dropout(x, keep_prob = p, is_training = self.isTraining, scope=scope)
					
					self.layerTable.append("dropout (p=%s)" % (p))
				else:
					assert False, "Invalid layer type '%s'" % type
					
			print("Network output shape: "+str(x.get_shape()))
			self.yUnactivated = x

			self.y = tf.sigmoid(self.yUnactivated, name="seqNetworkActivation")
			self.layerTable.append("sigmoid")
			self.yPred = tf.cast(tf.greater(self.y, 0.5), tf.float32)
			if config.classBalancing:
				classSizes = np.asfarray([annEnc.classSizes[annEnc.coordToNode[i]] for i in range(annEnc.getDim())])
				assert np.all(classSizes > 0)
				meanClassSize = np.mean(classSizes)
				posWeight = meanClassSize / classSizes
				# config.maxClassInbalance prevents too large effective learning rates (i.e. too large gradients)
				assert config.maxClassInbalance >= 1.0
				posWeight = np.maximum(1.0, np.minimum(config.maxClassInbalance, posWeight))
				self.loss = tf.nn.weighted_cross_entropy_with_logits(logits=self.yUnactivated, targets=self.yTarget, pos_weight=posWeight)
			else:
				self.loss = tf.nn.sigmoid_cross_entropy_with_logits(logits=self.yUnactivated, labels=self.yTarget)
			self.loss = tf.reduce_mean(self.loss, name="loss_all")
				
			# per-class statistics
			self.predPositivesClass = tf.reduce_sum(self.yPred, 0)
			self.actualPositivesClass = tf.reduce_sum(self.yTarget, 0)
			self.tpClass = tf.reduce_sum(self.yPred * self.yTarget, 0)

			# all-class statistics
			self.predPositives = tf.reduce_sum(self.predPositivesClass)
			self.actualPositives = tf.reduce_sum(self.actualPositivesClass)
			self.tp = tf.reduce_sum(self.tpClass)

			paramCount = networkParamCount(networkScope.name)
			print("Network has %s params (%.2f MB in float32)" % (paramCount, paramCount*4 / (1024.0*1024.0)))
			
	# Exports human-readable CSV description of the neural network architecture.
	def exportDescription(self, fnPrefix):
		exportTable(self.layerTable, fnPrefix, config.csvSeparator)
