#!/usr/bin/env python3

# SECLAF Sequence Classification Framework
# Copyright (c) 2016-2017 Balazs Szalkai

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

print("SECLAF Sequence Classification Framework -- Inference tool")
print("Copyright (c) 2016-2017 Balazs Szalkai")

import os,sys,re,math,time
print("Python "+sys.version)

import numpy as np
import tensorflow as tf
import json
import glob

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from seclaf import config, network
from seclaf import *

def log(s):
	print(s)

startTime = time.time()

ckptFilename = config.outFilePrefix+".ckpt"
assert ckptExists(ckptFilename), "Neural network checkpoint file does not exist"

inputFilename = sys.argv[2]
outputFilename = sys.argv[3] if len(sys.argv) >= 4 else None

assert os.path.isfile(inputFilename), "Sequence input file does not exist"
log("Loading input sequences from %s..." % inputFilename)
with open(inputFilename) as f:
	seqList = [line.strip() for line in f]
seqList = [s for s in seqList if s != ""]

# detect FASTA
isFasta = False
for s in seqList:
	if s.startswith(">") or s.startswith(";"):
		isFasta = True
		break
if isFasta:
	l2 = []
	currSeq = ""
	seqList.append(">") # to emit the last seq
	for line in seqList:
		if line.startswith(">") or line.startswith(";"):
			if currSeq != "":
				l2.append(currSeq.replace("*", "")) # there may be a * at the end of the sequence
			currSeq = ""
		else:
			currSeq += line
	seqList = l2
		
log("Successfully loaded %s sequences" % len(seqList))
assert len(seqList) > 0, "Error: zero sequences given"

tree = Tree(config.treeFile)
seqEncoder = config.seqEncoder()
classSizes = getClassSizesCached(tree, None, config.classSizesCacheFile)
annEnc = AnnotationEncoder(tree, config.maxLevel, config.minClassSize, classSizes)
config.maxLevel = annEnc.maxLevel

global_step = tf.Variable(tf.constant(0, dtype=tf.int32), trainable=False)
net = network.SeqNetwork(seqEncoder, annEnc)

def convertToBatch(seqList):
	inputSeq_value = []
	inputSeqLen_value = []

	for seq in seqList:
		if len(seq) > config.maxLength:
			print("Warning: truncated sequence %s..., length %s > %s" % (seq[:20], len(seq), config.maxLength))
			seq = seq[:config.maxLength]
		seqEncoded = seqEncoder.encode(seq)
		if len(seq) < config.maxLength:
			seqEncoded = np.pad(seqEncoded, ((0,config.maxLength-len(seq)),(0,0)), mode="constant")
		inputSeq_value.append(seqEncoded)
		inputSeqLen_value.append(len(seq))
	
	padCount = config.batchSize - len(seqList)
	if padCount > 0:
		inputSeq_value += [np.zeros((config.maxLength, seqEncoder.getDim()), dtype=np.float32)] * padCount
		inputSeqLen_value += [0] * padCount

	assert len(inputSeq_value) == config.batchSize
	inputSeq_value = np.stack(inputSeq_value, 0)
	inputSeqLen_value = np.stack(inputSeqLen_value, 0)
	return inputSeq_value, inputSeqLen_value

log("Creating session (tensorflow version: %s)" % tf.__version__)
sess_conf = tf.ConfigProto()
sess_conf.gpu_options.allow_growth = True
with tf.Session(config=sess_conf) as sess:
	log("Initializing variables")
	sess.run(tf.global_variables_initializer())
	log("Creating Saver")
	saver = tf.train.Saver()
	
	log("Loading checkpoint '%s'..." % ckptFilename)
	saver.restore(sess, os.path.abspath(ckptFilename)) # without abspath it crashes with FileNotFoundError
	log("Loading took %s seconds" % int(time.time() - startTime))
	log("Checkpoint was trained for %s steps" % sess.run(global_step))
	
	fout = open(outputFilename, "w") if outputFilename is not None else sys.stdout
	startTime = time.time()
	for i in range(int(math.ceil(float(len(seqList)) / config.batchSize))):
		seqListPart = seqList[i*config.batchSize : (i+1)*config.batchSize]
		if len(seqListPart) > 0:
			inputSeq_value, inputSeqLen_value = convertToBatch(seqListPart)
			y_value = sess.run(net.y, feed_dict={net.inputSeq:inputSeq_value, net.inputSeqLen:inputSeqLen_value, net.isTraining:False})

			for b in range(len(seqListPart)):
				classes, confidences = annEnc.decode(y_value[b])
				fout.write("@\t" + seqListPart[b] + "\t" + "&".join(classes) + "\t" + "&".join([str(q) for q in confidences])+"\n")
	if fout is not sys.stdout:
		fout.close()
				
	elapsed = time.time() - startTime
	log("Inference took %s seconds" % elapsed)

print("Exiting")
