#!/usr/bin/env python3

# SECLAF Sequence Classification Framework
# Copyright (c) 2016-2017 Balazs Szalkai

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

print("SECLAF Sequence Classification Framework -- Training tool")
print("Copyright (c) 2016-2017 Balazs Szalkai")

import os,re,math,time,sys
import numpy as np
import tensorflow as tf
import json
import glob
import sklearn.metrics
import matplotlib as mpl
mpl.use("Agg")
import matplotlib.pyplot as plt

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from seclaf import config, network
from seclaf import *

startTime = time.time()

assert os.path.exists(config.trainAnnotFile), config.trainAnnotFile + " does not exist"
assert os.path.exists(config.testAnnotFile), config.testAnnotFile + " does not exist"

ckptFilename = config.outFilePrefix+".ckpt"
checkpointExists = ckptExists(ckptFilename)
if checkpointExists:
	print("Last checkpoint %s will be loaded" % ckptFilename)
trainCsvFilename = config.outFilePrefix+".train.csv"

logFile = open(config.outFilePrefix+".log", "a")
def log(msg):
	print(msg)
	logFile.write(str(msg)+"\n")
	logFile.flush()
	
tree = Tree(config.treeFile)
tree.saveGraphml(config.treeFile+".graphml")
annTrain = AnnotationReader(config.trainAnnotFile, minLength=config.minLength, maxLength=config.maxLength, removeStartingM=config.removeStartingM)
annTest = AnnotationReader(config.testAnnotFile, minLength=config.minLength, maxLength=config.maxLength, removeStartingM=config.removeStartingM)
seqEncoder = config.seqEncoder()
classSizes = getClassSizesCached(tree, annTrain, config.classSizesCacheFile)
annTrain = AnnotationShuffler(annTrain)	# this should be done AFTER getting the class sizes!
annEnc = AnnotationEncoder(tree, config.maxLevel, config.minClassSize, classSizes)
config.maxLevel = annEnc.maxLevel

global_step = tf.Variable(tf.constant(0, dtype=tf.int32), trainable=False)

learning_rate = config.lrate
if isinstance(learning_rate, str) or isinstance(learning_rate, bytes):
	learning_rate = eval(learning_rate) # allow tensor expressions depending on global_step (i.e. lrate schedules)
learning_rate = tf.convert_to_tensor(learning_rate) # it may already be a Tensor because of eval
if config.lrateDecay < 1:
	learning_rate = tf.train.exponential_decay(learning_rate, global_step, config.lrateDecaySteps, config.lrateDecay, staircase=True)

net = network.SeqNetwork(seqEncoder, annEnc)
net.exportDescription(config.outFilePrefix + ".network")
regLoss = config.regLambda * tf.add_n(list(tf.get_collection(tf.GraphKeys.REGULARIZATION_LOSSES)))
net.loss += regLoss

# update ops are needed for batch normalization
update_ops = list(tf.get_collection(tf.GraphKeys.UPDATE_OPS))
with tf.control_dependencies(update_ops):
	optimizer = tf.train.AdamOptimizer(learning_rate=learning_rate).minimize(net.loss, global_step=global_step)

# Reads a new training batch.
# Returns: sequences(list), classes(list), encoded sequences (ndarray), lengths of the sequences (ndarray), encoded targets (ndarray).
def makeBatch(ann, rewind):
	seqList = []
	classList = []
	inputSeq_value = []
	inputSeqLen_value = []
	yTarget_value = []
	for i in range(config.batchSize):
		seq, classes = ann.read()
		if seq is None:
			if rewind:
				ann.rewind()
				seq, classes = ann.read()
			else:
				return None, None, None, None, None
		seqList.append(seq)
		classes = tree.includeParents(classes)
		classList.append(classes)
		seqEncoded = seqEncoder.encode(seq)
		if len(seq) < config.maxLength:
			seqEncoded = np.pad(seqEncoded, ((0,config.maxLength-len(seq)),(0,0)), mode="constant")
		inputSeq_value.append(seqEncoded)
		inputSeqLen_value.append(len(seq))
		yTarget_value.append(annEnc.encode(classes))
	inputSeq_value = np.stack(inputSeq_value, 0)
	inputSeqLen_value = np.stack(inputSeqLen_value, 0)
	yTarget_value = np.stack(yTarget_value, 0)
	return seqList, classList, inputSeq_value, inputSeqLen_value, yTarget_value

# Plots a simple line chart and saves it as png and svg.
def plotCurve(x, y, title, legend, xname, yname, fnPrefix):
	plt.ioff()
	fig = plt.figure()
	plt.title(title)
	plt.plot(x, y, color="red", lw=2, label=legend)
	plt.plot([0, 1], [0, 1], color="navy", lw=2, linestyle="--")
	plt.xlim([0.0, 1.0])
	plt.ylim([0.0, 1.05])
	plt.xlabel(xname)
	plt.ylabel(yname)
	plt.legend(loc="lower right")
	plt.savefig(fnPrefix+".svg")
	plt.savefig(fnPrefix+".png")
	plt.close(fig)
	
# Writes a ROC curve to a CSV file.
def writeCurveCsv(fn, header, data):
	table = np.stack(data, 1)
	with open(fn, "w") as f:
		f.write(toCsv(header, config.csvSeparator))
		for i in range(table.shape[0]):
			f.write(toCsv(table[i], config.csvSeparator))
	
log("Creating session (tensorflow version: %s)" % tf.__version__)
with tf.Session() as sess:
	log("Initializing variables")
	sess.run(tf.global_variables_initializer())
	log("Creating Saver")
	saver = tf.train.Saver()
	
	if checkpointExists:
		log("Loading checkpoint '%s'..." % ckptFilename)
		saver.restore(sess, os.path.abspath(ckptFilename)) # without abspath it crashes with FileNotFoundError
	else:
		log("Checkpoint not found, training from scratch")
		
	loss_sum = 0.0
	regLoss_sum = 0.0
	predPositives_sum = 0.0
	actualPositives_sum = 0.0
	tp_sum = 0.0
	loss_count = 0
	log("Loading took %s seconds" % int(time.time() - startTime))
	iterStart = sess.run(global_step)
	log("Training from iteration %s..." % iterStart)
	startTime = time.time()
	with open(trainCsvFilename, "w") as trainCsv:
		trainCsv.write(toCsv(["iter", "loss", "regLoss", "precision", "recall", "lrate", "elapsed"], config.csvSeparator))
		for iter in range(iterStart, config.iters+1):
			seqList_dummy, classList_dummy, inputSeq_value, inputSeqLen_value, yTarget_value = makeBatch(annTrain, True)
			
			optimizer_value_dummy, loss_value, regLoss_value, learning_rate_value, predPositives_value, actualPositives_value, tp_value = \
				sess.run([optimizer, net.loss, regLoss, learning_rate, net.predPositives, net.actualPositives, net.tp], \
					feed_dict={net.inputSeq:inputSeq_value, net.inputSeqLen:inputSeqLen_value, net.yTarget:yTarget_value, net.isTraining:True})
			
			loss_sum += loss_value
			regLoss_sum += regLoss_value
			predPositives_sum += predPositives_value
			actualPositives_sum += actualPositives_value
			tp_sum += tp_value
			loss_count += 1
			
			sys.stderr.write(".")
			sys.stderr.flush()
			
			if iter == config.iters or (iter % config.statStep == 0 and iter > 0):
				elapsed = time.time() - startTime
				secPerIter = elapsed / float(iter-iterStart + 1)
				precision = tp_sum / predPositives_sum
				recall = tp_sum / actualPositives_sum
				sys.stderr.write("\n")
				sys.stderr.flush()
				log("Iter %s, loss %.4g (of which regLoss %.4g), precision %.4f, recall %.4f, lrate %s, elapsed %s s, %.2f sec per iter" % 
					(iter, loss_sum/loss_count, regLoss_sum/loss_count, precision, recall, learning_rate_value, int(elapsed), secPerIter))
				trainCsv.write(toCsv([iter, loss_sum/loss_count, regLoss_sum/loss_count, precision, recall, learning_rate_value, elapsed], config.csvSeparator))
				trainCsv.flush()
					
				loss_sum = 0.0
				predPositives_sum = 0.0
				actualPositives_sum = 0.0
				tp_sum = 0.0
				loss_count = 0
				
			if config.iters > 1 and (iter == config.iters or (iter % config.saveStep == 0 and iter > 0)):
				log("Saving model to %s" % ckptFilename)
				saver.save(sess, ckptFilename, global_step=(iter if config.ckptKeep else None), write_meta_graph=False)
		
	log("Testing...")
	annTestOut = AnnotationWriter(config.testAnnotFile+".pred")
	testCount = 0
	predPositives_sum = 0.0
	actualPositives_sum = 0.0
	tp_sum = 0.0
	levelSuccessCount = [0] * (config.maxLevel+1)
	rocScore = []
	rocLabels = []
	for testIter in range(int(config.maxTestCount / config.batchSize)):
		seqList, classList, inputSeq_value, inputSeqLen_value, yTarget_value = makeBatch(annTest, False)
		if inputSeq_value is None:
			log("Test set exhausted after %s batches" % testIter)
			break
		y_value, yPred_value, predPositives_value, actualPositives_value, tp_value = sess.run([net.y, net.yPred, net.predPositivesClass, net.actualPositivesClass, net.tpClass],
			feed_dict={net.inputSeq:inputSeq_value, net.inputSeqLen:inputSeqLen_value, net.yTarget:yTarget_value, net.isTraining:False})
		sys.stderr.write("#")
		sys.stderr.flush()

		rocScore.append(y_value)
		rocLabels.append(yTarget_value)

		predPositives_sum += predPositives_value
		actualPositives_sum += actualPositives_value
		tp_sum += tp_value
		assert len(tp_sum) == annEnc.getDim()

		for b in range(config.batchSize):
			pred, _ = annEnc.decode(yPred_value[b])
			annTestOut.write(seqList[b], pred)
			
			testCount += 1
			for level in range(config.maxLevel+1):
				predNodes = [n for n in pred if annEnc.nodeLevel[n] == level]
				targetNodes = [n for n in classList[b] if n in annEnc.nodeLevel and annEnc.nodeLevel[n] == level]
				if set(predNodes) == set(targetNodes):
					levelSuccessCount[level] += 1
		
	levelSuccessRate = np.asfarray(levelSuccessCount) / float(testCount)
	precisionClass = tp_sum / np.maximum(1, predPositives_sum) # where predPositives_sum == 0, tp_sum is also 0
	recallClass = tp_sum / np.maximum(1, actualPositives_sum) # where actualPositives_sum == 0, tp_sum is also 0
	precision = np.sum(tp_sum) / np.sum(predPositives_sum)
	recall = np.sum(tp_sum) / np.sum(actualPositives_sum)
	f1 = 2*precision*recall / (precision + recall)
	log("Tested on %s seqs, precision %.2f%%, recall %.2f%%, F1 %.2f%%, perfect success rate per level is %s%%" % (testCount, 100*precision, 100*recall, 100*f1, 100*levelSuccessRate))
	
	rocScore = np.concatenate(rocScore, 0) # [testCount, n_classes]
	rocLabels = np.concatenate(rocLabels, 0) # [testCount, n_classes]
	assert rocScore.shape[0] == testCount
	assert rocScore.shape[1] == annEnc.getDim()
	
	# ROC curve and AUC
	fpr, tpr, thresholds = sklearn.metrics.roc_curve(y_true = np.reshape(rocLabels, (-1)), y_score = np.reshape(rocScore, (-1))) # micro-average ROC curve
	writeCurveCsv(fn = config.outFilePrefix+".roc.csv", header = ["fpr", "tpr", "threshold"], data = (fpr, tpr, thresholds))
	auc = sklearn.metrics.auc(fpr, tpr)
	log("AUC (with micro-averaging): %.5f" % auc)	
	plotCurve(x=fpr, y=tpr, title = config.name + " ROC curve", legend = config.shortName + " (AUC = %0.4f)" % auc,
		xname = "False positive rate", yname = "True positive rate", fnPrefix = config.outFilePrefix+".roc")

	# PR curve
	precisionPoints, recallPoints, thresholds = sklearn.metrics.precision_recall_curve(y_true = np.reshape(rocLabels, (-1)), probas_pred = np.reshape(rocScore, (-1))) # micro-average PR curve
	writeCurveCsv(fn = config.outFilePrefix+".pr.csv", header = ["recall", "precision"], data = (recallPoints, precisionPoints))
	plotCurve(x=recallPoints, y=precisionPoints, title = config.name + " PR curve", legend = config.shortName,
		xname = "Recall", yname = "Precision", fnPrefix = config.outFilePrefix+".pr")
		
	accuracyPerLevel = []
	accuracyPerLevel.append(["Level", "Classes", "Precision", "Recall", "F1-value"])
	for level in range(annEnc.maxLevel+1):
		a = annEnc.levelStart[level]
		b = annEnc.levelEnd[level]
		precisionLevel = np.sum(tp_sum[a:b]) / np.sum(predPositives_sum[a:b])
		recallLevel = np.sum(tp_sum[a:b]) / np.sum(actualPositives_sum[a:b])
		f1Level = 2*precisionLevel*recallLevel / (precisionLevel + recallLevel)
		accuracyPerLevel.append([level, b-a, "%.2f%%"%(100*precisionLevel), "%.2f%%"%(100*recallLevel), "%.2f%%"%(100*f1Level)])
	exportTable(accuracyPerLevel, config.outFilePrefix+".accuracy_per_level", config.csvSeparator)
		
	with open(config.outFilePrefix+".accuracy.csv", "w") as f:
		f.write(toCsv(["coord", "id", "name", "level", "precision", "recall", "classSize", "testClassSize"], config.csvSeparator))
		for i in range(annEnc.getDim()):
			coord = i
			node = annEnc.coordToNode[i]
			name = tree.nodeDesc[node]
			level = int(annEnc.nodeLevel[node])
			precision = precisionClass[i]
			recall = recallClass[i]
			classSize = int(classSizes[node])
			testClassSize = int(actualPositives_sum[i])
			f.write(toCsv([coord, node, name, level, precision, recall, classSize, testClassSize], config.csvSeparator))
	
	log("Testing done.")
	
logFile.close()
print("Exiting")
