#!/usr/bin/env python3

# SECLAF Sequence Classification Framework
# Copyright (c) 2016-2017 Balazs Szalkai

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os,sys,math,re,time
import numpy as np

# seq_util.py: Biological sequence utility functions.

# Our canonical ordering of the nucleotides.
Nucleotides = "AGCT"

# Gets the index of a nucleotide in our canonical ordering.
# returns <0 if x is invalid, or a number in 0..3 for any valid nucleotide letter
def getDnaIndex(x):
	i = Nucleotides.find(x)
	return i

# The opposite of getDnaIndex.
# x: 0..3, an index of a nucleotide.
def getDnaFromIndex(x):
	return Nucleotides[x]

# Alphabetical ordering of the amino acids.
AminoAcids = "ACDEFGHIKLMNPQRSTVWY"
assert len(AminoAcids) == 20

# Gets the index of an amino acid in our canonical ordering.
# returns <0 if x is invalid, or a number in 0..19 for any valid amino acid letter
def getAminoAcidIndex(x):
	i = AminoAcids.find(x)
	return i

# The opposite of getAminoAcidIndex.
# x: 0..19, an index of an amino acid.
def getAminoAcidFromIndex(x):
	return AminoAcids[x]

# Gets the expected charge of an amino acid (1: positive, -1: negative, 0: neutral).
# Since "H" is positive only 10% of the time, 0.1 is returned for "H".
def getAminoAcidCharge(x):
	if x in "KR":
		return 1.0
	if x == "H":
		return 0.1
	if x in "DE":
		return -1.0
	return 0.0
	
_hydro = [1.8, 2.5, -3.5, -3.5, 2.8, -0.4, -3.2, 4.5, -3.9, 3.8, 1.9, -3.5, -1.6, -3.5, -4.5, -0.8, -0.7, 4.2, -0.9, -1.3]
# Returns the hydrophobicity of an amino acid (identified by its letter).
def getAminoAcidHydrophobicity(x):
	return _hydro[getAminoAcidIndex(x)]
	
# Returns whether the amino acid (identified by its letter) is polar.
def isAminoAcidPolar(x):
	return x in "DEHKNQRSTY"
	
# Returns whether the amino acid (identified by its letter) is an aromatic compound.
def isAminoAcidAromatic(x):
	return x in "FWY"
	
# Returns whether the amino acid (identified by its letter) has a hydroxyl group.
def hasAminoAcidHydroxyl(x):
	return x in "ST"
	
# Returns whether the amino acid (identified by its letter) has a sulfur atom.
def hasAminoAcidSulfur(x):
	return x in "CM"
	
# Gets the molar weight of an amino acid (g).
def getAminoAcidMolarWeight(x):
	return [89.094, 121.154, 133.104, 147.131, 165.192, 75.067, 155.156, 131.175, 146.189, 131.175, 149.208, 132.119, 115.132, 146.146, 174.203, 105.093, 119.119, 117.148, 204.228, 181.191][getAminoAcidIndex(x)]
	
# Translates a single codon into the amino acid it encodes.
# x: a DNA string of length 3
# Returns: the amino acid letter, or "?" if x is an invalid or stop codon.
# "M" is returned for the start codon ATG.
def dnaTranslateSingle(x):
	if x in ["GCT", "GCC", "GCA", "GCG"]: return "A"
	if x in ["TTA", "TTG", "CTT", "CTC", "CTA", "CTG"]: return "L"
	if x in ["CGT", "CGC", "CGA", "CGG", "AGA", "AGG"]: return "R"
	if x in ["AAA", "AAG"]: return "K"
	if x in ["AAT", "AAC"]: return "N"
	if x in ["ATG"]: return "M"
	if x in ["GAT", "GAC"]: return "D"
	if x in ["TTT", "TTC"]: return "F"
	if x in ["TGT", "TGC"]: return "C"
	if x in ["CCT", "CCC", "CCA", "CCG"]: return "P"
	if x in ["CAA", "CAG"]: return "Q"
	if x in ["TCT", "TCC", "TCA", "TCG", "AGT", "AGC"]: return "S"
	if x in ["GAA", "GAG"]: return "E"
	if x in ["ACT", "ACC", "ACA", "ACG"]: return "T"
	if x in ["GGT", "GGC", "GGA", "GGG"]: return "G"
	if x in ["TGG"]: return "W"
	if x in ["CAT", "CAC"]: return "H"
	if x in ["TAT", "TAC"]: return "Y"
	if x in ["ATT", "ATC", "ATA"]: return "I"
	if x in ["GTT", "GTC", "GTA", "GTG"]: return "V"
	#if x in ["ATG"]: return ":" # start codon, but also M
	#if x in ["TAA", "TGA", "TAG"]: return "." # stop codon
	return "?"
	
# Translates a DNA string into an amino acid string.
# x: the DNA to translate, its length must be divisible by 3 (the codon length)
def dnaTranslate(x):
	result = ""
	assert len(x) % 3 == 0
	for i in range(0, len(x), 3):
		result += dnaTranslateSingle(x[i:i+3])
	return result
	
# Returns the number of different places between two sequences of the same length.
# (In other words, calculates the Hamming distance.)
def seqDiff(a, b):
	assert len(a) == len(b)
	result = 0
	for i in range(len(a)):
		if a[i] != b[i]:
			result += 1
	return result

# Returns the properly normalized difference of two angles in radians. Result: -pi..pi.
def angleDiffRad(diff):
	return np.remainder(diff + math.pi, 2*math.pi) - math.pi

# Returns the properly normalized difference of two angles in degrees. Result: -180..180 (-180..179 for integers).
def angleDiffDeg(diff):
	return np.remainder(diff + 180, 360) - 180
