#!/usr/bin/env python3

# SECLAF Sequence Classification Framework
# Copyright (c) 2016-2017 Balazs Szalkai

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os,sys,re,math,time
from .utils import *

# tree.py: Class for loading and saving class hierarchy trees.

# Class for loading and saving class hierarchy trees.
class Tree:
	# If filename is not None, the tree is loaded from the specified file.
	# Otherwise an empty tree is created.
	def __init__(self, filename=None):
		self.clear()
		if filename is not None:
			self.load(filename)
		
	# Resets the tree to the empty tree.
	def clear(self):
		self.parents = {}
		self.children = {}
		self.nodeDesc = {}
		self.roots = {}

	# Adds a node to the tree, identified by nodeId.
	# Does not create a new node if the node already exists.
	def addNode(self, nodeId):
		if nodeId not in self.parents:
			self.parents[nodeId] = {}
			self.children[nodeId] = {}
			self.nodeDesc[nodeId] = nodeId
			self.roots[nodeId] = True

	# Always adds a new node. If a node with the given id exists, appends a number at the end of the id to make it unique.
	# Returns the ID of the created node.
	def forceAddNode(self, nodeId):
		for i in range(len(self.parents)+10):
			id = nodeId
			if i > 0:
				id += "_"+str(i)
			if id not in self.parents:
				self.addNode(id)
				return id
		assert False
	
	# Deletes a node from the tree, identified by nodeId.
	def removeNode(self, nodeId):
		if nodeId in self.parents:
			del self.parents[nodeId]
			del self.children[nodeId]
			del self.nodeDesc[nodeId]
			isRoot = nodeId in self.roots
			if isRoot:
				del self.roots[nodeId]
			for otherNode, d in diter(self.parents):
				if nodeId in d:
					del d[nodeId]
					if len(d) == 0:
						self.roots[otherNode] = True
			for otherNode, d in diter(self.children):
				if nodeId in d:
					del d[nodeId]
			
	# Marks "parent" as a parent of "nodeId".
	# If they don't exist, adds the two nodes to the tree.
	def addParent(self, nodeId, parent):
		self.addNode(nodeId)
		self.addNode(parent)
		self.parents[nodeId][parent] = True
		self.children[parent][nodeId] = True
		if nodeId in self.roots:
			del self.roots[nodeId]
		
	# Returns a collection of the identifiers of all root nodes of the tree.
	def getRoots(self):
		return self.roots.keys()
		
	# Saves the tree in the own .tre format.
	# fn: the output filename (should end with .tre)
	def save(self, fn):
		print("Saving tree to "+fn)
		with open(fn, "w") as f:
			f.write("// Tree file for classification hierarchy\n")
			for node, parents in diter(self.parents):
				parents_text = "&".join(parents.keys())
				if parents_text == "":
					parents_text = "-"
				f.write(node+"\t"+parents_text+"\t"+self.nodeDesc[node].replace("\t", " ")+"\n")
	
	# Loads a tree from the own .tre format.
	# Each row must contain at most 3 tab-separated items: a node identifier, its parents separated by & (or a single hyphen if the node is a root), and the node description.
	# If there is no hierarchy (each class is a root), then each row may contain only the node identifier.
	# Comments may be included and have to start with //.
	def load(self, fn):
		self.clear()
		print("Loading tree from "+fn)
		with open(fn) as f:
			for line in f:
				line = line.rstrip()
				if line.startswith("//"):
					continue
				tok = line.split("\t")
				node = tok[0]
				parents = tok[1] if len(tok) >= 2 else "-"
				desc = tok[2] if len(tok) >= 3 else node
				if parents != "-":
					parents = parents.split("&")
					for p in parents:
						self.addParent(node, p)
				else:
					self.addNode(node)
				self.nodeDesc[node] = desc
		assert len(self.roots) > 0, "Tree has no root nodes"
		print("Tree has %s roots" % (len(self.roots)))
				
	# Makes sure that if a node is in nodeIds, then its all parents are there, too.
	# nodeIds: list of strings.
	# Returns: a copy of nodeIds with all the parents included (appended at the end).
	def includeParents(self, nodeIds):
		def recurse(n):
			for p in self.parents[n]:
				if p not in result:
					result.append(p)
					recurse(p)
		
		if isinstance(nodeIds, str):
			nodeIds = [nodeIds]
		result = list(nodeIds)
		for n in nodeIds: 
			recurse(n)
		return result
	
	# Internal helper function for saving a graphml file.
	def _saveGraphmlHelper(self, fn, parentRelation):
		sys.stdout.write("Saving tree as GraphML to "+fn+" ...")
		with open(fn, "w") as f:
			f.write("""<?xml version="1.0" encoding="UTF-8"?>
				<graphml xmlns="http://graphml.graphdrawing.org/xmlns"  
				xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
				xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns
				 http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd">""")

			keyId = "id"
			keyName = "name"
			f.write("<key id=\"d_%s\" for=\"node\" attr.name=\"%s\" attr.type=\"string\"/>" % (keyId, keyId))
			f.write("<key id=\"d_%s\" for=\"node\" attr.name=\"%s\" attr.type=\"string\"/>" % (keyName, keyName))

			f.write("<graph id=\"G\" edgedefault=\"directed\">\n")
			for node in parentRelation:
				f.write("<node id=\"%s\">\n" % htmlentities(node))
				f.write("<data key=\"d_%s\">%s</data>\n" % (keyId, htmlentities(node)))
				f.write("<data key=\"d_%s\">%s</data>\n" % (keyName, htmlentities(self.nodeDesc[node])))
				f.write("</node>\n")
			for node in parentRelation:
				for p in parentRelation[node]:
					f.write("<edge directed=\"true\" source=\"%s\" target=\"%s\"/>\n" % (htmlentities(node), htmlentities(p)))
					
			f.write("</graph>\n")
			f.write("</graphml>\n")
		print("Done")
		
	# Internal helper function.
	def _getSubtree(self, subtree, node, downward):
		subtree.addNode(node)
		for otherNode in (self.children if downward else self.parents)[node]:
			if downward:
				subtree.addParent(otherNode, node)
			else:
				subtree.addParent(node, otherNode)
			self._getSubtree(subtree, otherNode, downward)
		
	# Gets the subtree accessible from node (node will be a max. level node in this subtree).
	# The result will be a newly created Tree instance.
	# If downward is True, then instead includes all children and descendants of the node.
	def getSubtree(self, node, downward = False):
		subtree = Tree()
		self._getSubtree(subtree, node, downward)
		return subtree

	# Saves the tree as a graphml file.
	# If leafNode is not None, then saves only the subtree accessible from leafNode (leafNode and all its ancestors).
	def saveGraphml(self, fn, leafNode = None):
		if leafNode is not None:
			subtree = self.getSubtree(leafNode)
			parents = subtree.parents
		else:
			parents = self.parents
		self._saveGraphmlHelper(fn, parents)
