#!/usr/bin/env python3

# SECLAF Sequence Classification Framework
# Copyright (c) 2016-2017 Balazs Szalkai

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os,sys,re,math,time
import glob
import numpy as np
import tensorflow as tf

# utils.py: Various assorted utilities.

# Returns the iterable items of a dictionary. Python 2/3 compatibility function.
def diter(d):
	try:
		return d.iteritems()
	except:
		return d.items()
		
# Converts a list ("a") to a csv row, using the given separator (, or ;).
def toCsv(a, sep):
	if isinstance(a, str):
		return a
		
	if isinstance(a, np.ndarray):
		assert a.ndim == 1
		a = a.tolist()
		
	if sep==";":
		return sep.join([(str(x).replace(".", ",") if isinstance(x, float) or isinstance(x, np.ndarray) else str(x)) for x in a])+"\n"
	else:
		return sep.join([str(x) for x in a])+"\n"

# Returns whether a Tensorflow checkpoint of the given name exists.
def ckptExists(ckptFilename):
	return len(glob.glob(ckptFilename+".data-*")) > 0
	
# Escapes html entities in the given string, for writing it into a html file.
def htmlentities(s):
	return s.replace("&", "&amp;").replace('"', "&quot;").replace("<", "&lt;").replace(">", "&gt;")

# Converts rows (a list of lists or a list of strings) into a HTML table representation.
# rows may contain HTML markup. Not sanitized.
# Returns the result as a single string.
def getHtmlTable(rows):
	return "<table border='1'>\n<tr><td>" + \
		"</td></tr>\n<tr><td>".join(["</td><td>".join([str(x) for x in row]) if isinstance(row, list) else str(row) for row in rows]) + \
		"</td></tr>\n</table>\n"
	
# Escapes s for LaTeX output.
def texEscape(s):
	if not isinstance(s, str):
		s = str(s)
	s = s.replace("%", "\\%").replace("#", "\\#")
	# TODO not all characters are escaped
	return s
	
# Converts rows (a list of lists or a list of strings) into a LaTeX table representation.
# rows may contain HTML markup, which will be converted to LaTeX markup in a limited sense.
def getTexTable(rows):
	return "\\begin{tabular}{|c|}\n\\hline\n" + \
		"".join([(" & ".join([texEscape(x) for x in row]) if isinstance(row, list) else texEscape(row)) + " \\\\ \n\\hline\n" for row in rows]).replace("<b>", "\\textbf{").replace("</b>", "}") + \
		"\\end{tabular}\n"
		
# Converts rows (a list of lists or a list of strings) into a CSV table representation.
# rows may contain HTML markup. Not sanitized.
def getCsvTable(rows, sep):
	return "".join([toCsv(x, sep) for x in rows]).replace("<b>", "").replace("</b>", "")
	
# Exports rows (a list of lists or a list of strings) as a table.
# Exports html, tex and csv files as well, appending the extension to fnPrefix to get the output filenames.
def exportTable(rows, fnPrefix, sep):
	with open(fnPrefix+".html", "w") as f:
		f.write(getHtmlTable(rows))
	with open(fnPrefix+".tex", "w") as f:
		f.write(getTexTable(rows))
	with open(fnPrefix+".csv", "w") as f:
		f.write(getCsvTable(rows, sep))
